
# Imports {{{1
# ==============================================================================
suppressMessages(library(adegenet))
suppressMessages(library(ggplot2))
suppressMessages(library(RColorBrewer))
suppressMessages(library(scales)) # for `trans_*`, including `trans_new`
# }}}1

# Plotting {{{1

## Support {{{2

# Multiple plot function
# From: http://www.cookbook-r.com/Graphs/Multiple_graphs_on_one_page_(ggplot2)/
#
# ggplot objects can be passed in ..., or to plotlist (as a list of ggplot objects)
# - cols:   Number of columns in layout
# - layout: A matrix specifying the layout. If present, 'cols' is ignored.
#
# If the layout is something like matrix(c(1,2,3,3), nrow=2, byrow=TRUE),
# then plot 1 will go in the upper left, 2 will go in the upper right, and
# 3 will go all the way across the bottom.
#
multiplot <- function(..., plotlist=NULL, file, cols=1, layout=NULL) {
  library(grid)

  # Make a list from the ... arguments and plotlist
  plots <- c(list(...), plotlist)

  numPlots = length(plots)

  # If layout is NULL, then use 'cols' to determine layout
  if (is.null(layout)) {
    # Make the panel
    # ncol: Number of columns of plots
    # nrow: Number of rows needed, calculated from # of cols
    layout <- matrix(seq(1, cols * ceiling(numPlots/cols)),
                    ncol = cols, nrow = ceiling(numPlots/cols))
  }

 if (numPlots==1) {
    print(plots[[1]])

  } else {
    # Set up the page
    grid.newpage()
    pushViewport(viewport(layout = grid.layout(nrow(layout), ncol(layout))))

    # Make each plot, in the correct location
    for (i in 1:numPlots) {
      # Get the i,j matrix positions of the regions that contain this subplot
      matchidx <- as.data.frame(which(layout == i, arr.ind = TRUE))

      print(plots[[i]], vp = viewport(layout.pos.row = matchidx$row,
                                      layout.pos.col = matchidx$col))
    }
  }
}

# }}}2

## Performance Over Parameter Space {{{2

# plotPerformanceOverParameterSpace  {{{3
# This plot produces a tile plot, with trait transition rates and
# dispersal rates on the the x and y axes, respectively, and performance
# (measured as proportion of correctly classified models) indicated by tile
# color.
#
# NOTE: A crucial assumption is that all parameters have been sampled
# at a limited number of points. That is, while the birth rate, for example, is
# a continuous parameter, it has be sampled at relatively small number of
# points, and the same goes for all other parameters (death rate, dispersal
# rate, trait transition rate, etc.). This is typical for one approach to
# parameter space exploration with simulations. The other approach, where
# parameters are sampled from a range, will not work here. You will need to bin
# the paramters before passing the data to this function.
#
# This functions requires as input a data.frame() object with the following
# fields:
#
#   'true.model.proportion.correctly.assigned'
#   'true.model.posterior.mean'
#   'birth.rate'
#   'death.rate'
#   'dispersal.rate'
#   'trait.transition.rate'
#
# This is obtained by running:
#
#       $ archipelago-classify \
#           --true-model=<model> \
#           -l birth.date:<b> \
#           -l death.rate:<e> \
#           -l dispersal.rate:<d> \
#           -l trait.transition.rate <q> \
#           <target-summary-statistics.csv> \
#           <training-summary-statistics.csv> <training-summary-statistics.csv> ...
#
# If `facet.by.diversification.rates` is True, then the plot will be faceted by
# birth and death rates.
#
# Note that the axes will be log-scaled, because it makes sense to me to
# explore parameter space that way.
#
plotPerformanceOverParameterSpace <- function(performance.df,
                                              response.metric="true.model.proportion.correctly.assigned",
                                              facet.by.diversification.rates=T
                                              ) {

    if (response.metric == "true.model.proportion.correctly.assigned") {
        response.metric.desc = "Proportion Correct"
    } else if (response.metric == "true.model.posterior.mean") {
        response.metric.desc = "Mean Posterior of True Model"
    } else {
        stop(paste("Unrecognized response metric: '", response.metric, "'", sep=""))
    }

    performance.breaks = c(0.0, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    # performance.breaks = c(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    f1 <- cut(
            performance.df[[response.metric]],
            breaks=performance.breaks,
            right=F,
            )
    performance.df$response.metric.factor <- factor(f1, levels=levels(f1))
    performance.df$birth.rate.factor <- factor(paste("b = ", performance.df$birth.rate, sep=""))
    performance.df$death.rate.factor <- factor(paste("e = ",
                                                        ifelse(performance.df$death.rate==0, "0", paste(format(round(performance.df$death.rate/performance.df$birth.rate, 2), nsmall=2), " x b", sep="")),
                                                        sep=""
                                                        ))

    performance.df$x.val = round(log10(performance.df$dispersal.rate/performance.df$birth.rate), 2)
    performance.df$x = factor(performance.df$x)
    performance.df$y.val = round(log10(performance.df$trait.transition.rate/performance.df$birth.rate), 2)
    performance.df$y = factor(performance.df$y)

    # drop extreme values: in earlier simulations, an unreasonably broad part of parameter space was explored
    performance.df = subset(performance.df, log10(performance.df$dispersal.rate/performance.df$birth.rate) >= -1 & log10(performance.df$trait.transition.rate/performance.df$birth.rate) >= -1)

    xlabels =  paste("b x 10E", sprintf("%+1.2f", performance.df$x.val), sep="")
    names(xlabels) <- performance.df$x
    ylabels =  paste("b x 10E", sprintf("%+1.2f", performance.df$y.val), sep="")
    names(ylabels) <- performance.df$y


    p <- ggplot(performance.df, aes(x, y))

    p <- p + geom_tile(aes(fill=response.metric.factor))
    p <- p + scale_x_discrete(labels=xlabels)
    p <- p + scale_y_discrete(labels=ylabels)
    # palette = "Greys"
    # palette = "YlGn"
    palette = "RdYlBu"
    if (length(performance.breaks) > 9) {
        color.fn <- colorRampPalette(brewer.pal(9, palette))
        p <- p + scale_fill_manual(values=color.fn(length(performance.breaks)), name=response.metric.desc)
    } else {
        p <- p + scale_fill_brewer(palette=palette, name=response.metric.desc)
    }

    p <- p + labs(x = "Dispersal Rate (in Units of Birth Rate)", y = "Trait Transition Rate (in Units of Birth Rate)")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )

    if (facet.by.diversification.rates) {
        if (length(levels(performance.df$death.rate.factor)) > 1 && length(levels(performance.df$death.rate.factor)) > 1) {
            # p <- p + facet_grid(birth.rate.factor ~ death.rate.factor, labeller= label_parsed)
            p <- p + facet_grid(birth.rate.factor ~ death.rate.factor)
        } else if (length(levels(performance.df$death.rate.factor)) > 1) {
            p <- p + facet_wrap(~death.rate.factor)
        } else if (length(levels(performance.df$birth.rate.factor)) > 1) {
            p <- p + facet_wrap(~birth.rate.factor)
        }
    }

    return(p)
}
# }}}3

plotPerformanceOverContinuousParameterSpace <- function(performance.df,
                                              response.metric="true.model.proportion.correctly.assigned",
                                              facet.by.diversification.rates=T
                                              ) {

    if (response.metric == "true.model.proportion.correctly.assigned") {
        response.metric.desc = "Proportion Correct"
    } else if (response.metric == "true.model.posterior.mean") {
        response.metric.desc = "Mean Posterior of True Model"
    } else {
        stop(paste("Unrecognized response metric: '", response.metric, "'", sep=""))
    }

    # performance.breaks = c(0.0, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    performance.breaks = c(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    f1 <- cut(
            performance.df[[response.metric]],
            breaks=performance.breaks,
            right=F,
            )
    performance.df$response.metric.factor <- factor(f1, levels=levels(f1))
    performance.df$birth.rate.factor <- factor(paste("b = ", performance.df$birth.rate, sep=""))
    performance.df$death.rate.factor <- factor(paste("e = ",
                                                        ifelse(performance.df$death.rate==0, "0", paste(format(round(performance.df$death.rate/performance.df$birth.rate, 2), nsmall=2), " x b", sep="")),
                                                        sep=""
                                                        ))

    performance.df$x.val = round(log10(performance.df$dispersal.rate/performance.df$birth.rate), 2)
    performance.df$y.val = round(log10(performance.df$trait.transition.rate/performance.df$birth.rate), 2)

    # drop extreme values: in earlier simulations, an unreasonably broad part of parameter space was explored
    performance.df = subset(performance.df, log10(performance.df$dispersal.rate/performance.df$birth.rate) >= -1 & log10(performance.df$trait.transition.rate/performance.df$birth.rate) >= -1)

    xlabels =  unique(paste("b x 10E", sprintf("%+1.2f", performance.df$x.val), sep=""))
    names(xlabels) <- unique(performance.df$x.val)
    ylabels =  unique(paste("b x 10E", sprintf("%+1.2f", performance.df$y.val), sep=""))
    names(ylabels) <- unique(performance.df$y.val)

    p <- ggplot(performance.df, aes(x.val, y.val))
    # p <- p + geom_point(aes(color=response.metric.factor))
    p <- p + geom_tile(aes(fill=response.metric.factor))
    p <- p + scale_x_continuous(breaks=unique(performance.df$x.val), labels=xlabels)
    p <- p + scale_y_continuous(breaks=unique(performance.df$y.val), labels=ylabels)
    # palette = "Greys"
    palette = "YlGn"
    # palette = "RdYlBu"
    if (length(performance.breaks) > 9) {
        color.fn <- colorRampPalette(brewer.pal(9, palette))
        p <- p + scale_fill_manual(values=color.fn(length(performance.breaks)), name=response.metric.desc)
    } else {
        p <- p + scale_fill_brewer(palette=palette, name=response.metric.desc)
    }

    p <- p + labs(x = "Dispersal Rate (in Units of Birth Rate)", y = "Trait Transition Rate (in Units of Birth Rate)")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )

    if (facet.by.diversification.rates) {
        if (length(levels(performance.df$death.rate.factor)) > 1 && length(levels(performance.df$death.rate.factor)) > 1) {
            # p <- p + facet_grid(birth.rate.factor ~ death.rate.factor, labeller= label_parsed)
            p <- p + facet_grid(birth.rate.factor ~ death.rate.factor)
        } else if (length(levels(performance.df$death.rate.factor)) > 1) {
            p <- p + facet_wrap(~death.rate.factor)
        } else if (length(levels(performance.df$birth.rate.factor)) > 1) {
            p <- p + facet_wrap(~birth.rate.factor)
        }
    }

    empirical.df = data.frame(dispersal=log10(0.0179814/0.0548), trait.transition.rate=log10(0.062247/0.0548))
    p <- p + geom_point(aes(x=log10(0.0179814/0.0548),y=log10(0.062247/0.0548)))

    return(p)
}
# }}}3

plotPerformanceOverParameterSpaceDispersalRateAgainstBirthRateByTraitTransitionRate <- function(performance.df) { # {{{3
    performance.breaks = c(0.0, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    f1 <- cut(
            performance.df[["true.model.proportion.correctly.assigned"]],
            breaks=performance.breaks,
            right=F,
            )
    performance.df$true.model.proportion.correctly.assigned.factor <- factor(f1, levels=levels(f1))
    performance.df$x.val = round(log10(performance.df$dispersal.rate/performance.df$birth.rate), 2)
    performance.df$x = factor(performance.df$x.val)
    performance.df$y.val = round(log10(performance.df$birth.rate), 2)
    performance.df$y = factor(performance.df$y.val)
    performance.df$dispersal.vs.transition.rate <- factor(ifelse(performance.df$dispersal.rate > performance.df$trait.transition.rate, "dispersal rate > trait transition rate",
                                                                 ifelse(performance.df$dispersal.rate < performance.df$trait.transition.rate, "dispersal rate < trait transition rate", "dispersal rate = trait transition rate")))
    # performance.df$dispersal.vs.transition.rate <- factor(performance.df$dispersal.rate/performance.df$trait.transition.rate)
    # drop extreme values: in earlier simulations, an unreasonably broad part of parameter space was explored
    performance.df = subset(performance.df, log10(performance.df$dispersal.rate/performance.df$birth.rate) >= -1 & log10(performance.df$trait.transition.rate/performance.df$birth.rate) >= -1)
    xlabels =  paste("b x 10E", sprintf("%+1.2f", performance.df$x.val), sep="")
    names(xlabels) <- performance.df$x
    ylabels =  paste(performance.df$birth.rate)
    names(ylabels) <- performance.df$y
    p <- ggplot(performance.df, aes(x, y))
    p <- p + geom_tile(aes(fill=true.model.proportion.correctly.assigned.factor))
    p <- p + scale_x_discrete(labels=xlabels)
    p <- p + scale_y_discrete(labels=ylabels)
    palette = "RdYlBu"
    if (length(performance.breaks) > 9) {
        color.fn <- colorRampPalette(brewer.pal(9, palette))
        p <- p + scale_fill_manual(values=color.fn(length(breaks)), name="Proportion Correct")
    } else {
        # p <- p + scale_fill_brewer(type="seq", palette=palette, name="")
        p <- p + scale_fill_brewer(palette=palette, name="Proportion Correct")
    }
    p <- p + labs(x = "Dispersal Rate (in Units of Birth Rate)", y = "Birth Rate")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    p <- p + facet_grid(.~dispersal.vs.transition.rate)
    return(p)
}
# }}}3

plotPerformanceOverParameterSpaceDispersalRateAgainstTraitTransitionRateByBirthRate <- function(performance.df) { # {{{3
    performance.breaks = c(0.0, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    f1 <- cut(
            performance.df[["true.model.proportion.correctly.assigned"]],
            breaks=performance.breaks,
            right=F,
            )
    performance.df$true.model.proportion.correctly.assigned.factor <- factor(f1, levels=levels(f1))
    performance.df$x.val = round(log10(performance.df$dispersal.rate/performance.df$birth.rate), 2)
    performance.df$x = factor(performance.df$x.val)
    performance.df$y.val = round(log10(performance.df$trait.transition.rate/performance.df$birth.rate), 2)
    performance.df$y = factor(performance.df$y.val)
    performance.df$dispersal.rate.relative.to.birth.rate <- factor(ifelse(performance.df$dispersal.rate > performance.df$birth.rate, "dispersal rate > birth rate",
                                                                 ifelse(performance.df$dispersal.rate < performance.df$birth.rate, "dispersal rate < birth rate", "dispersal rate = birth rate")))
    performance.df$trait.transition.rate.relative.to.birth.rate <- factor(ifelse(performance.df$trait.transition.rate > performance.df$birth.rate, "trait.transition rate > birth rate",
                                                                 ifelse(performance.df$trait.transition.rate < performance.df$birth.rate, "trait.transition rate < birth rate", "trait.transition rate = birth rate")))

    # drop extreme values: in earlier simulations, an unreasonably broad part of parameter space was explored
    performance.df = subset(performance.df, log10(performance.df$dispersal.rate/performance.df$birth.rate) >= -1 & log10(performance.df$trait.transition.rate/performance.df$birth.rate) >= -1)

    xlabels =  paste("b x 10E", sprintf("%+1.2f", performance.df$x.val), sep="")
    names(xlabels) <- performance.df$x
    ylabels =  paste("b x 10E", sprintf("%+1.2f", performance.df$y.val), sep="")
    names(ylabels) <- performance.df$y

    p <- ggplot(performance.df, aes(x, y))

    p <- p + geom_tile(aes(fill=true.model.proportion.correctly.assigned.factor))
    p <- p + scale_x_discrete(labels=xlabels)
    p <- p + scale_y_discrete(labels=ylabels)
    palette = "RdYlBu"
    if (length(performance.breaks) > 9) {
        color.fn <- colorRampPalette(brewer.pal(9, palette))
        p <- p + scale_fill_manual(values=color.fn(length(breaks)), name="Proportion Correct")
    } else {
        p <- p + scale_fill_brewer(palette=palette, name="Proportion Correct")
    }

    p <- p + labs(x = "Dispersal Rate (in Units of Birth Rate)", y = "Trait Transition Rate (in Units of Birth Rate)")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    p <- p + facet_grid(dispersal.rate.relative.to.birth.rate ~ trait.transition.rate.relative.to.birth.rate)
    return(p)
}
# }}}3

# (Performance Over Parameter Space) }}}2

## Posterior Probability of True/False Model Over Parameter Space {{{2

plotPosteriorsOverParameterSpace <- function(classification.df, performance.df=NULL) { # {{{3
    # classification.df = classification.df[classification.df$true.model==true.model,]
    df1 = classification.df
    df1$response.group = "Posterior of True Model"
    df1$response.metric = df1$true.model.posterior
    df2 = classification.df
    df2$response.group = "Posterior of False Model"
    df2$response.metric = 1 - df2$true.model.posterior
    combined.classification.df = rbind(df1, df2)
    # combined.classification.df$response.metric.factor <- ifelse( combined.classification.df$response.metric >= 0.99, "[0.99,1.00]",
    #                                                    ifelse(   combined.classification.df$response.metric >= 0.95, "[0.95,0.99)",
    #                                                    ifelse(   combined.classification.df$response.metric >= 0.90, "[0.90,0.95)",
    #                                                    ifelse(   combined.classification.df$response.metric >= 0.50, "[0.50,0.90)",
    #                                                    ifelse(   combined.classification.df$response.metric > 0.10,  "(0.10,0.50)",
    #                                                    ifelse(   combined.classification.df$response.metric > 0.05,  "(0.05,0.10]",
    #                                                    ifelse(   combined.classification.df$response.metric > 0.01,  "(0.01,0.05]",
    #                                                                                                                           "[0.00,0.01]"
    #                                                          )))))))
    combined.classification.df$response.metric.factor <- ifelse( combined.classification.df$response.metric >= 0.99, "0.99-1.00",
                                                       ifelse(   combined.classification.df$response.metric >= 0.95, "0.95-0.99",
                                                       ifelse(   combined.classification.df$response.metric >= 0.90, "0.90-0.95",
                                                       ifelse(   combined.classification.df$response.metric >= 0.50, "0.50-0.90",
                                                       ifelse(   combined.classification.df$response.metric > 0.10,  "0.10-0.50",
                                                       ifelse(   combined.classification.df$response.metric > 0.05,  "0.05-0.10",
                                                       ifelse(   combined.classification.df$response.metric > 0.01,  "0.01-0.05",
                                                                                                                     "0.00-0.01"
                                                             )))))))
    combined.classification.df$x.val = round(log10(combined.classification.df$dispersal.rate / combined.classification.df$trait.transition.rate), 2)
    combined.classification.df$x = factor(combined.classification.df$x.val)
    combined.classification.df$facet.group = paste('True Model="', combined.classification.df$true.model, '"', sep="")
    # xlabels =  paste("x10E", sprintf("%+1.2f", combined.classification.df$x.val, sep=""))
    # names(xlabels) <- unique(combined.classification.df$x)
    xlabels =  c(
                 "x10E-2.00",
                 "x10E-1.75",
                 "x10E-1.50",
                 "x10E-1.25",
                 "x10E-1.00",
                 "x10E-0.75",
                 "x10E-0.50",
                 "x10E-0.25",
                 "x10E+0.00",
                 "x10E+0.25",
                 "x10E+0.50",
                 "x10E+0.75",
                 "x10E+1.00",
                 "x10E+1.25",
                 "x10E+1.50",
                 "x10E+1.75",
                 "x10E+2.00"
                 )
    p <- ggplot(combined.classification.df, aes(x, fill=response.metric.factor))
    p <- p + scale_x_discrete(labels=unique(xlabels))
    p <- p + geom_bar(position="fill")
    # palette = "Greys"
    # palette = "YlGn"
    palette = "RdYlBu"
    p <- p + scale_fill_brewer(palette=palette, name="")
    if (!is.null(performance.df )) {
        p.df1 = performance.df
        p.df1$x2 = factor(round(log10(p.df1$dispersal.rate / p.df1$trait.transition.rate), 2))
        p.df1$y2 = p.df1$true.model.proportion.correctly.assigned
        p.df1$response.group = "Posterior of True Model"
        p.df2 = performance.df
        p.df2$x2 = factor(round(log10(p.df2$dispersal.rate / p.df2$trait.transition.rate), 2))
        p.df2$y2 = 1-p.df1$true.model.proportion.correctly.assigned
        p.df2$response.group = "Posterior of False Model"
        combined.performance.df = rbind(p.df1, p.df2)
        p <- p + geom_point(data=combined.performance.df, aes(x2, y2, fill=""))
    }
    p <- p + facet_grid(facet.group~response.group)
    p <- p + labs(x = "Dispersal Rate (in Units of Trait Transition Rate)", y = "Proportion of Analyses")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  legend.text=element_text(size=8),
                  legend.title=element_text(size=8),
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    return(p)
}
# }}}3

plotHighPosteriorsOverParameterSpace <- function(classification.df, focus.on.true.model=T) { # {{{3
    # classification.df = classification.df[classification.df$true.model==true.model,]
    if (focus.on.true.model) {
        response.metric = "true.model.posterior"
        response.metric.desc = "Posterior of True Model"
    } else {
        classification.df$false.model.posterior = 1 - classification.df$true.model.posterior
        response.metric = "false.model.posterior"
        response.metric.desc = "Posterior of False Model"
    }
    # performance.breaks = c(0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0)
    # performance.breaks = c(0.0, 0.01, 0.05, 0.1, 0.5, 0.9, 0.95, 0.99, 1.1)
    # f1 <- cut(
    #         classification.df[[response.metric]],
    #         breaks=performance.breaks,
    #         right=F,
    #         )
    # classification.df$response.metric.factor <- factor(f1, levels=levels(f1))
    classification.df$response.metric.factor <- ifelse(        classification.df[[response.metric]] >= 0.99, "0.99-1.00",
                                                       ifelse( classification.df[[response.metric]] >= 0.95, "0.95-0.99",
                                                       ifelse( classification.df[[response.metric]] >= 0.90, "0.90-0.95",
                                                       ifelse( classification.df[[response.metric]] >= 0.50, "0.50-0.90",
                                                       ifelse( classification.df[[response.metric]] > 0.10,  "0.10-0.50",
                                                       ifelse( classification.df[[response.metric]] > 0.05,  "0.05-0.10",
                                                       ifelse( classification.df[[response.metric]] > 0.01,  "0.01-0.05",
                                                                                                             "0.00-0.01"
                                                             )))))))
    classification.df$x.val = round(log10(classification.df$dispersal.rate / classification.df$trait.transition.rate), 2)
    classification.df$x = factor(classification.df$x.val)
    classification.df$facet.group = paste('True Model="', classification.df$true.model, '"', sep="")
    xlabels =  paste("x 10E", sprintf("%+1.2f", classification.df$x.val, sep=""))
    # xlabels =  c(
    #              "x10E-2.00",
    #              "x10E-1.75",
    #              "x10E-1.50",
    #              "x10E-1.25",
    #              "x10E-0.00",
    #              "x10E-0.00",
    #              "x10E-0.75",
    #              "x10E-0.50",
    #              "x10E-0.25",
    #              "x10E+0.00",
    #              "x10E+0.25",
    #              "x10E+0.50",
    #              "x10E+0.75",
    #              "x10E+1.25",
    #              "x10E+1.50",
    #              "x10E+1.75",
    #              "x10E+2.00"
    #              )
    names(xlabels) <- classification.df$x
    p <- ggplot(classification.df, aes(x, fill=response.metric.factor))
    p <- p + scale_x_discrete(labels=unique(xlabels))
    p <- p + geom_bar(position="fill")
    # palette = "Greys"
    # palette = "YlGn"
    palette = "RdYlBu"
    p <- p + scale_fill_brewer(palette=palette, name=response.metric.desc)
    p <- p + facet_grid(.~facet.group)
    p <- p + labs(x = "Dispersal Rate (in Units of Trait Transition Rate)", y = "Proportion of Analyses")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  legend.text=element_text(size=7),
                  legend.title=element_text(size=6),
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    return(p)
}
# }}}3

# (Posterior Probability of True/False Model Over Parameter Space) }}}2

## Profiles {{{2
plotPostSimulationParameterEstimates <- function(profile.df, response.metric) { # {{{3
    profile.df$true.birth.rate <- profile.df$lineage.birth.rate.definition
    profile.df$est.birth.rate <- profile.df$pure.birth.rate
    profile.df$true.birth.rate.factor <- factor(profile.df$true.birth.rate)
    profile.df$scaled.true.trait.transition.rate <- profile.df$trait.habitat.true.transition.rate / profile.df$true.birth.rate
    profile.df$scaled.true.trait.transition.rate.factor = factor(profile.df$scaled.true.trait.transition.rate)
    profile.df$scaled.true.dispersal.rate <- profile.df$global.dispersal.rate / profile.df$true.birth.rate
    profile.df$scaled.true.dispersal.rate.factor = factor(profile.df$scaled.true.dispersal.rate)
    profile.df$scaled.est.trait.transition.rate <- profile.df$trait.habitat.est.transition.rate / profile.df$true.birth.rate
    profile.df$scaled.est.dispersal.rate <- profile.df$biogeobears.dec.dispersal.rate / profile.df$true.birth.rate
    if (response.metric == "dispersal") {
        # profile.df$x = log10(profile.df$global.dispersal.rate / profile.df$true.birth.rate)
        profile.df$x = log10(profile.df$global.dispersal.rate/profile.df$true.birth.rate)
        profile.df$x.factor = factor(profile.df$x)
        # profile.df$x.factor = factor(paste("b x 10E", sprintf("%+1.2f", log10(profile.df$global.dispersal.rate/profile.df$true.birth.rate)), sep=""))
        profile.df$y = log10(profile.df$biogeobears.dec.dispersal.rate/profile.df$true.birth.rate)

        profile.df = profile.df[log10(profile.df$biogeobears.dec.dispersal.rate) > -8,] # crude outlier removal
        xlabels =  paste("b x 10E", sprintf("%+1.2f", log10(profile.df$global.dispersal.rate/profile.df$true.birth.rate)), sep="")
        names(xlabels) <- profile.df$x.factor
        ylabels =  paste("b x 10E", sprintf("%+1.2f", unique(profile.df$x)), sep="")

        p <- ggplot(profile.df, aes(x.factor, y, fill=model.category))
        p <- p + geom_violin()
        p <- p + scale_y_continuous(breaks=unique(profile.df$x), labels=ylabels)

        # profile.df$ymin = min(profile.df[profile.df$x.factor,]$y)
        # profile.df$ymax = min(profile.df[profile.df$x.factor,]$y)
        # p <- ggplot(profile.df, aes(x, y, ymin=ymin, ymax=ymax, fill=model.category, color=model.category))
        # p <- p + geom_pointrange()

        p <- p + scale_x_discrete(labels=xlabels)
        p <- p + labs(x = "Simulation Dispersal Rate (in Units of Birth Rate)", y = "Estimated Dispersal Rate (in Units of Birth Rate)")
        # rotate labels
        p <- p + theme( axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
    } else if (response.metric == "dispersal-faceted-by-transition") {
        # profile.df$x = log10(profile.df$global.dispersal.rate / profile.df$true.birth.rate)
        profile.df$x = log10(profile.df$global.dispersal.rate/profile.df$true.birth.rate)
        profile.df$x.factor = factor(profile.df$x)
        # profile.df$x.factor = factor(paste("b x 10E", sprintf("%+1.2f", log10(profile.df$global.dispersal.rate/profile.df$true.birth.rate)), sep=""))
        profile.df$y = log10(profile.df$biogeobears.dec.dispersal.rate/profile.df$true.birth.rate)

        profile.df = profile.df[log10(profile.df$biogeobears.dec.dispersal.rate) > -8,] # crude outlier removal
        profile.df$facet = factor(log10(profile.df$scaled.true.trait.transition.rate))

        p <- ggplot(profile.df, aes(x.factor, y, fill=model.category, color=model.category))
        p <- p + geom_boxplot()
        p <- p + scale_y_continuous(limits=c(-3, 2))

        p <- p + labs(x = "Simulation Dispersal Rate Scaled by Birth Rate", y = "Estimated Dispersal Rate Scaled by Birth Rate")
        p <- p + theme( axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) )
        p <- p + facet_wrap(~facet)
    } else if (response.metric == "dispersal-scaled-by-transition") {
        scaling.factor = profile.df$scaled.true.trait.transition.rate
        profile.df$x = log10(profile.df$scaled.true.dispersal.rate / scaling.factor)
        profile.df$x.factor = factor(profile.df$x)
        profile.df$y = log10(profile.df$scaled.est.dispersal.rate / scaling.factor)
        profile.df = profile.df[log10(profile.df$biogeobears.dec.dispersal.rate) > -8,] # crude outlier removal
        xlabels =  paste("q x 10E", sprintf("%+1.2f", profile.df$x), sep="")
        names(xlabels) <- profile.df$x.factor
        ylabels =  paste("q x 10E", sprintf("%+1.2f", unique(profile.df$x)), sep="")
        p <- ggplot(profile.df, aes(x.factor, y, fill=model.category))
        p <- p + geom_violin(removePanelGrid=TRUE)
        p <- p + scale_y_continuous(breaks=unique(profile.df$x), labels=ylabels)
        p <- p + scale_x_discrete(labels=xlabels)
        p <- p + labs(x = "True Dispersal Rate as Multiple of Trait Transition Rate", y = "Estimated Dispersal Rate as Multiple of Transition Rate")
        p <- p + theme( axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
    # } else if (response.metric == "dispersal-scaled-by-transition") {
        # scaling.factor = profile.df$scaled.true.trait.transition.rate
        # profile.df$x = log10(profile.df$scaled.true.dispersal.rate / scaling.factor)
        # profile.df$x.factor = factor(profile.df$x)
        # profile.df$y = log10(profile.df$scaled.est.dispersal.rate / scaling.factor)
        # profile.df = profile.df[log10(profile.df$biogeobears.dec.dispersal.rate) > -8,] # crude outlier removal
        # p <- ggplot(profile.df, aes(x.factor, y, fill=model.category))
        # p <- p + labs(x = "True Dispersal Rate Scaled by True Trait Transition Rate", y = "Estimated Dispersal Rate Scaled By Trait Transition Rate")
        # p <- p + geom_boxplot()
    # } else if (response.metric == "dispersal-scaled-by-transition") {
    #     scaling.factor = profile.df$true.trait.transition.rate
    #     profile.df$x = log10(profile.df$global.dispersal.rate / scaling.factor)
    #     profile.df$x.factor = factor(profile.df$x)
    #     profile.df$y = log10(profile.df$biogeobears.dec.dispersal.rate / scaling.factor)
    #     profile.df = profile.df[log10(profile.df$biogeobears.dec.dispersal.rate) > -8,] # crude outlier removal
    #     xlabels =  paste("t x 10E", sprintf("%+1.2f", profile.df$x), sep="")
    #     names(xlabels) <- profile.df$x.factor
    #     ylabels =  paste("t x 10E", sprintf("%+1.2f", unique(profile.df$x)), sep="")
    #     p <- ggplot(profile.df, aes(x.factor, y, fill=model.category))
    #     p <- p + geom_violin()
    #     p <- p + scale_y_continuous(breaks=unique(profile.df$x), labels=ylabels)
    #     p <- p + scale_x_discrete(labels=xlabels)
    #     p <- p + labs(x = "True Dispersal Rate Scaled by True Trait Transition Rate", y = "Estimated Dispersal Rate Scaled By Trait Transition Rate")
    #     p <- p + theme( axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
    } else if (response.metric == "dispersal2") {
        p <- ggplot(profile.df, aes(scaled.true.dispersal.rate.factor, scaled.est.dispersal.rate, fill=model.category))
        p <- p + geom_boxplot()
        # p <- p + facet_wrap(~scaled.true.trait.transition.rate.factor)
    } else if (response.metric=="dispersal3") {
        scaling.factor = profile.df$scaled.true.trait.transition.rate
        profile.df$x = profile.df$global.dispersal.rate / scaling.factor
        profile.df$y = profile.df$biogeobears.dec.dispersal.rate
        profile.df = profile.df[which(profile.df$x <= 0.4), ]
        p <- ggplot(profile.df, aes(model.category, y))
        p <- p + geom_boxplot()
    } else if (response.metric=="trait") {
        profile.df$x = log10(profile.df$trait.habitat.true.transition.rate/profile.df$true.birth.rate)
        profile.df$x.factor = factor(profile.df$x)
        profile.df$y = log10(profile.df$trait.habitat.est.transition.rate/profile.df$true.birth.rate)

        xlabels =  paste("b x 10E", sprintf("%+1.2f", log10(profile.df$trait.habitat.true.transition.rate/profile.df$true.birth.rate)), sep="")
        names(xlabels) <- profile.df$x.factor
        ylabels =  paste("b x 10E", sprintf("%+1.2f", unique(profile.df$x)), sep="")

        p <- ggplot(profile.df, aes(x.factor, y, fill=model.category))
        p <- p + geom_violin()
        p <- p + scale_y_continuous(breaks=unique(profile.df$x), labels=ylabels)
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
        p <- p + scale_x_discrete(labels=xlabels)
        p <- p + labs(x = "Simulation Trait Transition Rate (in Units of Birth Rate)", y = "Estimated Trait Transition Rate (in Units of Birth Rate)")
        p <- p + theme( axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1))
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
    } else if (response.metric=="death") {
        p <- ggplot(profile.df, aes(lineage.death.rate.definition, biogeobears.dec.extinction.rate, fill=model.category))
        p <- p + geom_point()
        p <- p + labs(x = "Simulated Death Rate",
                      y = "Estimated Death Rate")
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
    } else if (response.metric=="birth") {
        ybreaks = unique(profile.df$true.birth.rate)
        p <- ggplot(profile.df, aes(true.birth.rate.factor, est.birth.rate, fill=model.category))
        p <- p + geom_violin()
        p <- p + scale_y_log10(breaks=ybreaks)
        p <- p + labs(x = "Simulated Birth Rate",
                      y = "Estimated Birth Rate")
        p <- p + guides(fill=guide_legend(title="Dispersal Regime"))
    } else {
        stop(paste("Unrecognized response metric: '", response.metric, "'", sep=""))
    }
    # p <- p + geom_violin()
    p <- p + theme(legend.position="bottom")
    return(p)
}
# }}}3

# (Profiles) }}}2

## Robustness {{{2

plotPosteriorRobustness <- function(classification.df, performance.df=NULL) { # {{{3
    misspecification.target = "birth"
    # classification.df = classification.df[classification.df$misspecification.target==misspecification.target,]
    df1 = classification.df
    classification.df$x.val = classification.df$misspecification.exponent
    classification.df$x = factor(classification.df$misspecification.exponent)
    classification.df$facet.group = paste('True Model="', classification.df$true.model, '"', sep="")
    # classification.df$misspecification.target.desc = paste('Parameter
    xlabels =  paste("x 10E", sprintf("%+1.2f", classification.df$x.val, sep=""))
    names(xlabels) <- classification.df$x.val
    classification.df$response.metric.factor <- ifelse( classification.df$true.model.posterior >= 0.99, "0.99-1.00",
                                                       ifelse(   classification.df$true.model.posterior >= 0.95, "0.95-0.99",
                                                       ifelse(   classification.df$true.model.posterior >= 0.90, "0.90-0.95",
                                                       ifelse(   classification.df$true.model.posterior >= 0.50, "0.50-0.90",
                                                       ifelse(   classification.df$true.model.posterior > 0.10,  "0.10-0.50",
                                                       ifelse(   classification.df$true.model.posterior > 0.05,  "0.05-0.10",
                                                       ifelse(   classification.df$true.model.posterior > 0.01,  "0.01-0.05",
                                                                                                                     "0.00-0.01"
                                                             )))))))
    p <- ggplot(classification.df, aes(x, fill=response.metric.factor))
    p <- p + scale_x_discrete(labels=unique(xlabels))
    # p <- p + scale_x_discrete(labels=unique(xlabels), limits=levels(xlabels))
    p <- p + geom_bar(position="fill")
    palette = "RdYlBu"
    p <- p + scale_fill_brewer(palette=palette, name="")
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  legend.text=element_text(size=8),
                  legend.title=element_text(size=8),
                  axis.text.x = element_text(size=8, angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    p <- p + labs(x = "Error in Parameter Value Calibration (in Units of True Value)", y = "Proportion of Analyses")
    if (!is.null(performance.df )) {
        performance.df$x2 = factor(performance.df$misspecification.exponent)
        p <- p + geom_point(data=performance.df, aes(x2, y=true.model.proportion.correctly.assigned, fill=""))
    }
    p <- p + geom_vline(xintercept = 9)
    p <- p + facet_grid(facet.group~misspecification.target)
    return(p)
}

# }}}3

plotSelectionRobustness <- function(performance.df, misspecification.target) { # {{{3
    df = performance.df[performance.df$misspecification.target==misspecification.target,]
    # df = performance.df
    df$x = df$misspecification.exponent
    df$y = df$true.model.proportion.correctly.assigned
    p <- ggplot(df, aes(x, y, color=true.model))
    p <- p + geom_point()
    return(p)
    p <- p + facet_grid(true.model~misspecification.target)
    p <- p + theme(
                  # panel.border=element_rect(size=0.5, fill=NA, color="black", linetype='dotted'),
                  legend.position="bottom",
                  legend.text=element_text(size=8),
                  legend.title=element_text(size=8),
                  axis.text.x = element_text(angle = 90, vjust = 0.5, hjust=1) # rotate x-axis labels
                  # axis.text.y = element_text(size=5)
                  )
    return(p)
}
# }}}3

# (Robustness}}}2

# (Plotting) }}}1

# Support Functions {{{1
# ==============================================================================
calcRMSE <- function(error) {
    return(sqrt(mean(error^2)))
}

calcMAE <- function(error) {
    return(mean(abs(error)))
}
# }}}1

# Higher-Level Convenience {{{1
# ==============================================================================

generatePerformancePlots <- function(performance.df, output.prefix="plots/performance-over-parameter-space") { # {{{3
    pdf(paste(output.prefix, "-unified.pdf", sep=""))
    print(plotPerformanceOverParameterSpace(performance.df, facet.by.diversification.rates=F))
    dev.off()
    pdf(paste(output.prefix, "-faceted-by-diversification.pdf", sep=""))
    print(plotPerformanceOverParameterSpace(performance.df, facet.by.diversification.rates=T))
    dev.off()
    pdf(paste(output.prefix, "-dispersal-against-birth-rate-by-transition-rate.pdf", sep=""))
    print(plotPerformanceOverParameterSpaceDispersalRateAgainstBirthRateByTraitTransitionRate(performance.df))
    dev.off()
    pdf(paste(output.prefix, "-dispersal-against-transition-rate-by-birth-rate.pdf", sep=""))
    print(plotPerformanceOverParameterSpaceDispersalRateAgainstTraitTransitionRateByBirthRate(performance.df))
    dev.off()
} # }}}3

generatePostSimulationParameterEstimates <- function(profile.df, output.prefix="plots/post-simulation-parameters-by-regime") { # {{{3
    pdf(paste(output.prefix, "-birth-rate.pdf", sep=""))
    print(plotPostSimulationParameterEstimates(profile.df, "birth"))
    dev.off()
    pdf(paste(output.prefix, "-dispersal-rate-scaled-by-transition-rate.pdf", sep=""))
    print(plotPostSimulationParameterEstimates(profile.df, "dispersal-scaled-by-transition"))
    dev.off()
    pdf(paste(output.prefix, "-dispersal-rate.pdf", sep=""))
    print(plotPostSimulationParameterEstimates(profile.df, "dispersal"))
    dev.off()
    pdf(paste(output.prefix, "-trait-transition-rate.pdf", sep=""))
    print(plotPostSimulationParameterEstimates(profile.df, "trait"))
    dev.off()
}
# }}}3


# }}}1
