#! /usr/bin/env python

import re
import sys
import os
import argparse
import collections
import csv
import re

from archipelago import model
sys.path.insert(0, os.path.dirname(__file__))
import validatelib

def main():
    """
    Main CLI handler.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("dirs", metavar="DIR", type=str, nargs="+", help="path to directories")
    parser.add_argument("-l", "--labels",
            action="append",
            help="labels to append to output (in format <FIELD-NAME>:value;)")
    parser.add_argument("-f", "--model-format",
            dest="model_file_schema",
            choices=["json", "python"],
            default="json",
            help="Format of model file.")
    parser.add_argument( "--no-header-row",
            action="store_true",
            default=False,
            help="Do not write a header row.")
    parser.add_argument( "--append",
            action="store_true",
            default=False,
            help="Append to output file if it already exists instead of overwriting.")
    args = parser.parse_args()
    extra_fields = validatelib.parse_fieldname_and_value(args.labels)
    sources = validatelib.get_source_filepaths(
            dirs=args.dirs,
            model_file_schema=args.model_file_schema)
    param_fieldnames = [
            "num.tips",
            "root.age",
            "num.focal.areas",
            "num.supplemental.areas",
            "birth.rate",
    ]
    param_fieldname_mapping = {
            "birth.rate": "lineage.birth.rate.definition",
    }
    output_fieldnames = list(extra_fields.keys())
    output_fieldnames.extend(param_fieldnames)
    output_fieldnames.extend([
            "trait.nstates",
            "trait.true.rate",
            "trait.est.rate",
    ])
    nstates_pattern = re.compile(r".*nst(\d+).*")
    out = sys.stdout
    results = []
    for model_filepath, csv_filepath in sources:
        archipelago_model = model.ArchipelagoModel.from_path(
                filepath=model_filepath,
                schema=args.model_file_schema)
        reader = csv.DictReader(open(csv_filepath, "r"))
        source_fieldnames = set(reader.fieldnames)

        true_transition_rate_source_fieldname = list(f for f in source_fieldnames if "true.transition.rate" in f)
        assert len(true_transition_rate_source_fieldname) == 1
        true_transition_rate_source_fieldname = true_transition_rate_source_fieldname[0]

        est_transition_rate_source_fieldname = list(f for f in source_fieldnames if "est.transition.rate" in f)
        assert len(est_transition_rate_source_fieldname) == 1
        est_transition_rate_source_fieldname = est_transition_rate_source_fieldname[0]

        for row in reader:
            record = collections.OrderedDict()
            record.update(extra_fields)
            for param_fieldname in param_fieldnames:
                record[param_fieldname] = row[param_fieldname_mapping.get(param_fieldname, param_fieldname)]
            nstates_match = nstates_pattern.match(true_transition_rate_source_fieldname)
            assert nstates_match
            record["trait.nstates"] = nstates_match.groups()[0]
            record["trait.true.rate"] = row[true_transition_rate_source_fieldname]
            record["trait.est.rate"] = row[est_transition_rate_source_fieldname]
            results.append(record)
    writer = csv.DictWriter(out,
            fieldnames=output_fieldnames)
    if not args.no_header_row:
        writer.writeheader()
    writer.writerows(results)

if __name__ == '__main__':
    main()


