#! /usr/bin/env python

import sys
import os
import json
import random
import argparse
import itertools
import collections

from archipelago import simulate

general_job_template = """\
#! /bin/bash
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -l h_vmem=12G
#$ -l virtual_free=12G
{queue}
{job_commands}
"""

def to_json(o):
    if isinstance(o, DynamicValue):
        return o.value
    raise TypeError(repr(o) + " is not JSON serializable")

class DynamicValue(object):

    def __init__(self, value=None):
        self.value = value

def generate_area_definitions(
        num_focal_areas,
        num_supplemental_areas=0):
    areas = []
    for i in range(num_supplemental_areas):
        areas.append({
            "label": "b{}".format(i+1),
            "is_supplemental": True,
            })
    for i in range(num_focal_areas):
        areas.append({
            "label": "a{}".format(i+1),
            "is_supplemental": False,
            })
    return areas

def generate_model(
        num_focal_areas,
        num_supplemental_areas,
        target_focal_area_lineages,
        birth_rate,
        death_rate,
        trait_transition_rate,
        global_dispersal_rate,
        unconstrained_dispersal_weight,
        constrained_dispersal_weight):
    model = {
        "areas" : generate_area_definitions(num_focal_areas=num_focal_areas, num_supplemental_areas=num_supplemental_areas),
        "diversification": {
            "lineage_birth_rate": {
                    "definition_type": "fixed_value",
                    "definition": birth_rate,
            },
            "lineage_death_rate": {
                    "definition_type": "fixed_value",
                    "definition": death_rate,
            },
        },
        "traits": [
            {"label": "habitat", "nstates": 2, "transition_rate": trait_transition_rate,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:habitat",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
        "termination_conditions": {"target_focal_area_lineages": target_focal_area_lineages},
    }
    return model

def get_summary_command(
        nreps,
        output_prefix,
        model_filepath,
        labels,
        rng):
    commands = []
    sim_cmd = []
    sim_cmd.append("archipelago-simulate.py")
    sim_cmd.extend(["--nreps", str(nreps)])
    sim_cmd.extend(["-z", str(rng.randint(0, sys.maxsize))])
    sim_cmd.extend(["--output-prefix", output_prefix])
    sim_cmd.extend(["--stderr-logging-level", "warning"])
    sim_cmd.append(os.path.basename(model_filepath))
    sim_cmd = " ".join(sim_cmd)
    commands.append(sim_cmd)
    trees_filepath = simulate.ArchipelagoSimulator.compose_focal_areas_trees_filepath(output_prefix)
    summary_path = "{}.summary.csv".format(output_prefix)
    summary_cmd = []
    summary_cmd.append("archipelago-summarize.py")
    for label in labels:
        summary_cmd.extend(["-l", label])
    summary_cmd.extend(["-o", summary_path])
    summary_cmd.append("-q")
    summary_cmd.append(trees_filepath)
    summary_cmd = " ".join(summary_cmd)
    commands.append(summary_cmd)
    commands = "\n".join(commands)
    return commands, summary_path

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-test-data-per-regime", "--ntest",
            type=int,
            default=30,
            help="Number of test data replicates per 'constrained' and 'unconstrained' regimes (default = %(default)s).")
    parser.add_argument("--num-training-data-per-regime", "--ntraining",
            type=int,
            default=100,
            help="Number of training data replicates per 'constrained' and 'unconstrained' regimes (default = %(default)s).")
    parser.add_argument(
            "--njob-replicates",
            "--num-replicates-per-configuration",
            type=int,
            default=1,
            help="Number of replicates per parameter combination (default = %(default)s).")
    parser.add_argument("--no-summary",
            action="store_true",
            default=False,
            help="Do not create summaries.")
    parser.add_argument("--profile", "--create-profile",
            dest="create_profile",
            action="store_true",
            default=False,
            help="Create profile (estimate birth, death, and dispersal rates on tree).")
    parser.add_argument("--no-profile-dispersal",
            action="store_true",
            default=False,
            help="Do not profile dispersal, even if profile is requested.")
    parser.add_argument("--venv",
            default=None,
            help="Path to Python virtual environment.")
    parser.add_argument("-z", "--random-seed",
            default=None,
            help="Seed for random number generator engine.")
    parser.add_argument("-q", "--queue",
            default=None,
            help="Name of queue to use (default: no queue specification")
    regime_options = parser.add_argument_group("Simulation Regime")
    regime_options.add_argument("--num-focal-areas",
            type=int,
            default=4,
            help="Number of focal areas (default: %(default)s).")
    regime_options.add_argument("--num-supplemental-areas",
            type=int,
            default=1,
            help="Number of supplemental areas (default: %(default)s).")
    regime_options.add_argument("--num-taxa",
            dest="target_focal_area_lineages",
            type=int,
            default=200,
            help="Number of taxa (default: %(default)s).")
    model_classification_options = parser.add_argument_group("Classification Options")
    model_classification_options.add_argument("--set-npca",
            default=None,
            metavar="n",
            help="Number of axes retained in the principal component analysis step.",
            )
    model_classification_options.add_argument("--optimize-npca",
            default=None,
            metavar="PENALTY-FACTOR",
            help=(
                "Use number of principal component axes that maximizes a "
                "score given by: (proportion of correct classifications  "
                "when reapplied to the training data) minus (the         "
                "proportion of available PC axes used multipled by the   "
                "given penalty factor). A penalty factor of 0.0 means to "
                "use the number of PC axes that maximizes the proportion "
                "of correct classifications. A penalty factor of 1.0     "
                "means that each additional axis retained will deduct    "
                "1.0/|PC| from the proportion of correct classification "
                "score."
                )
            )
    model_classification_options.add_argument("--maximize-npca",
            action="store_true",
            default=None,
            help="Retain all the principal component axes for the DAPC function.")

    args = parser.parse_args()

    if args.random_seed is None:
        args.random_seed = random.randint(0, sys.maxsize)
    rng = random.Random(args.random_seed)
    if args.venv is not None:
        venv_dir = os.path.expanduser(os.path.expandvars(args.venv))
        venv_activate = os.path.abspath(os.path.join(venv_dir, "bin", "activate"))
        if not os.path.exists(venv_activate):
            raise Exception("Virtual environment activation script not found: '{}'".format(venv_activate))
        source_venv = "source {}".format(venv_activate)
    else:
        source_venv = ""
    if args.queue:
        queue = "#$ -q {}".format(args.queue)
    else:
        queue = ""

    if not args.no_summary:
        num_chosen = len(list(itertools.ifilter(lambda x: x is not None, [args.set_npca, args.optimize_npca, args.maximize_npca])))
        if num_chosen == 0:
            sys.exit("Must specify one of: '--set-npca', '--optimize-npca', or '--maximize-npca'")
        elif num_chosen > 1:
            sys.exit("Cannot specify more than one of: '--set-npca', '--optimize-npca', or '--maximize-npca'")
        elif args.set_npca is not None:
            npca = ["--set-npca", args.set_npca]
        elif args.optimize_npca is not None:
            npca = ["--optimize-npca", args.optimize_npca]
        elif args.maximize_npca is not None:
            npca = ["--maximize-npca"]

    target_regimes = [
        #   b,    e,       d,     t,
        (0.030, 0.000, 0.030, 0.030),
            ]
    # misspecification_regimes = (-1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1)
    misspecification_regimes = (-2, -1.75, -1.5, -1.25, -1, -0.75, -0.5, -0.25, 0, 0.25, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2)

    run_idx = 0
    job_count = 0
    job_filepaths = set()
    target_model_count = 0
    target_model_names = set()
    training_model_count = 0
    training_model_names = set()
    expected_output_filenames = set()
    expected_output_filename_count = 0

    for target_regime_idx, target_regime in enumerate(target_regimes):
        true_birth_rate = target_regime[0]
        true_death_rate = target_regime[1]
        true_dispersal_rate = target_regime[2]
        true_trait_transition_rate = target_regime[3]
        for is_constrained in (True, False):
            if is_constrained:
                true_model_category = "constrained"
                true_constrained_dispersal_weight = 0.0
                true_unconstrained_dispersal_weight = 1.0 * 2
            else:
                true_model_category = "unconstrained"
                true_constrained_dispersal_weight = 1.0
                true_unconstrained_dispersal_weight = 1.0
            target_model = generate_model(
                    num_focal_areas=args.num_focal_areas,
                    num_supplemental_areas=args.num_supplemental_areas,
                    target_focal_area_lineages=args.target_focal_area_lineages,
                    birth_rate=true_birth_rate,
                    death_rate=true_death_rate,
                    trait_transition_rate=true_trait_transition_rate,
                    global_dispersal_rate=true_dispersal_rate,
                    unconstrained_dispersal_weight=true_unconstrained_dispersal_weight,
                    constrained_dispersal_weight=true_constrained_dispersal_weight,
                    )
            target_model_count += 1
            target_model_filepath = "c{:04d}_target_{}.json".format(target_model_count, true_model_category)
            target_model_names.add(target_model_filepath)
            with open(target_model_filepath, "w") as f:
                json.dump(target_model, f, indent=4)
            for misspecification_target in ("birth", "dispersal", "trait"):
                for misspecification_regime in misspecification_regimes:
                    if misspecification_target == "birth":
                        training_birth_rate = true_birth_rate * pow(10, misspecification_regime)
                    else:
                        training_birth_rate = true_birth_rate
                    if misspecification_target == "death":
                        training_death_rate = true_death_rate * pow(10, misspecification_regime)
                    else:
                        training_death_rate = true_death_rate
                    if misspecification_target == "trait_transition":
                        training_trait_transition_rate = true_trait_transition_rate * pow(10, misspecification_regime)
                    else:
                        training_trait_transition_rate = true_trait_transition_rate
                    if misspecification_target == "dispersal":
                        training_dispersal_rate = true_dispersal_rate * pow(10, misspecification_regime)
                    else:
                        training_dispersal_rate = true_dispersal_rate
                    unconstrained_training_model = generate_model(
                            num_focal_areas=args.num_focal_areas,
                            num_supplemental_areas=args.num_supplemental_areas,
                            target_focal_area_lineages=args.target_focal_area_lineages,
                            birth_rate=training_birth_rate,
                            death_rate=training_death_rate,
                            trait_transition_rate=training_trait_transition_rate,
                            global_dispersal_rate=training_dispersal_rate,
                            unconstrained_dispersal_weight=1.0,
                            constrained_dispersal_weight=1.0,
                            )
                    unconstrained_training_model_filepath = "c{:04d}_training_{}_x_10E{}_unconstrained.json".format(target_model_count, misspecification_target, misspecification_regime)
                    training_model_count += 1
                    training_model_names.add(unconstrained_training_model_filepath)
                    with open(unconstrained_training_model_filepath, "w") as f:
                        json.dump(unconstrained_training_model, f, indent=4)
                    constrained_training_model = generate_model(
                            num_focal_areas=args.num_focal_areas,
                            num_supplemental_areas=args.num_supplemental_areas,
                            target_focal_area_lineages=args.target_focal_area_lineages,
                            birth_rate=training_birth_rate,
                            death_rate=training_death_rate,
                            trait_transition_rate=training_trait_transition_rate,
                            global_dispersal_rate=training_dispersal_rate,
                            unconstrained_dispersal_weight=2.0,
                            constrained_dispersal_weight=0.0,
                            )
                    constrained_training_model_filepath = "c{:04d}_training_{}_x_10E{}_constrained.json".format(target_model_count, misspecification_target, misspecification_regime)
                    training_model_count += 1
                    training_model_names.add(constrained_training_model_filepath)
                    with open(constrained_training_model_filepath, "w") as f:
                        json.dump(constrained_training_model, f, indent=4)
                    for job_idx in range(args.njob_replicates):
                        run_idx += 1
                        run_prefix = "r{:05d}".format(run_idx)
                        job_commands = []
                        target_summary_commands, target_summary_path = get_summary_command(
                                nreps=args.num_test_data_per_regime,
                                output_prefix=run_prefix + ".target",
                                model_filepath=target_model_filepath,
                                labels=[],
                                rng=rng)
                        job_commands.append(target_summary_commands)
                        expected_output_filename_count += 1
                        expected_output_filenames.add(target_summary_path)
                        unconstrained_training_summary_commands, unconstrained_training_summary_path = get_summary_command(
                                nreps=args.num_training_data_per_regime,
                                output_prefix=run_prefix + ".training-unconstrained",
                                model_filepath=unconstrained_training_model_filepath,
                                labels=["model.category:unconstrained"],
                                rng=rng)
                        job_commands.append(unconstrained_training_summary_commands)
                        expected_output_filename_count += 1
                        expected_output_filenames.add(unconstrained_training_summary_path)
                        constrained_training_summary_commands, constrained_training_summary_path = get_summary_command(
                                nreps=args.num_training_data_per_regime,
                                output_prefix=run_prefix + ".training-constrained",
                                model_filepath=constrained_training_model_filepath,
                                labels=["model.category:constrained"],
                                rng=rng)
                        job_commands.append(constrained_training_summary_commands)
                        expected_output_filename_count += 1
                        expected_output_filenames.add(constrained_training_summary_path)

                        model_classification_output_name = "{}.classification.csv".format(run_prefix)
                        expected_output_filename_count += 1
                        expected_output_filenames.add(model_classification_output_name)
                        performance_output_name = "{}.performance.csv".format(run_prefix)
                        expected_output_filename_count += 1
                        expected_output_filenames.add(performance_output_name)
                        performance_command = []
                        performance_command.append("archipelago-classify.py")
                        performance_command.extend(npca)
                        performance_command.extend(["-l", "true.model:{}".format(true_model_category)])
                        performance_command.extend(["-l", "true.birth.rate:{}".format(true_birth_rate)])
                        performance_command.extend(["-l", "true.death.rate:{}".format(true_death_rate)])
                        performance_command.extend(["-l", "true.dispersal.rate:{}".format(true_dispersal_rate)])
                        performance_command.extend(["-l", "true.trait.transition.rate:{}".format(true_trait_transition_rate)])
                        performance_command.extend(["-l", "training.birth.rate:{}".format(training_birth_rate)])
                        performance_command.extend(["-l", "training.death.rate:{}".format(training_death_rate)])
                        performance_command.extend(["-l", "training.dispersal.rate:{}".format(training_dispersal_rate)])
                        performance_command.extend(["-l", "training.trait.transition.rate:{}".format(training_trait_transition_rate)])
                        performance_command.extend(["-l", "misspecification.target:{}".format(misspecification_target)])
                        performance_command.extend(["-l", "misspecification.exponent:{}".format(misspecification_regime)])
                        performance_command.extend(["-l", "misspecification.factor:{}".format(pow(10, misspecification_regime))])
                        performance_command.extend(["-l", "num.focal.areas:{}".format(args.num_focal_areas)])
                        performance_command.extend(["-l", "num.supplemental.areas:{}".format(args.num_supplemental_areas)])
                        performance_command.extend(["-l", "ntax:{}".format(args.target_focal_area_lineages)])
                        performance_command.extend(["--true-model", true_model_category])
                        performance_command.extend(["--primary-output-filepath", model_classification_output_name])
                        performance_command.extend(["--performance-output-filepath", performance_output_name])
                        performance_command.append(target_summary_path)
                        performance_command.append(unconstrained_training_summary_path)
                        performance_command.append(constrained_training_summary_path)
                        performance_command = " ".join(performance_command)
                        job_commands.append(performance_command)

                        job_filename = "{}.job".format(run_prefix)
                        job_count += 1
                        sys.stderr.write("Job: {}\n".format(job_filename))
                        with open(job_filename, "w") as jobf:
                            template = general_job_template
                            jobf.write(template.format(
                                job_commands="\n".join(job_commands),
                                queue=queue,
                                ))
                        # job_filepaths.append((subdir, job_filename))
                        job_filepaths.add(job_filename)

    assert len(job_filepaths) == job_count
    assert len(target_model_names) == target_model_count
    assert len(training_model_names) == training_model_count
    assert len(expected_output_filenames) == expected_output_filename_count
    with open("submit.sh", "w") as out:
        out.write("#! /bin/bash\n")
        for job_filepath in job_filepaths:
            out.write("qsub {}\n".format(job_filepath))
    sys.stderr.write("---\n")
    sys.stderr.write("{} jobs created\n".format(len(job_filepaths)))

if __name__ == "__main__":
    main()


