#! /usr/bin/env python

import sys
import os
import json
import random
import argparse
import itertools
import collections
from archipelago import profile

sys.path.insert(0, os.path.dirname(__file__))
import validatelib

def main():
    parser = argparse.ArgumentParser(
            parents=[
                validatelib.JobBuilder.get_general_validator_arg_parser(),
                ],
            )
    parser.add_argument("--estimate-dec-biogeobears",
            action="store_true",
            default=False,
            help="Estimate parameters under Lagrange's DEC model (using BioGeoBears).")
    parser.add_argument("--estimate-dec-lagrange",
            action="store_true",
            default=False,
            help="Estimate parameters under Lagrange's DEC model (using Lagrange).")
    parser.add_argument("--estimate-area-transition",
            action="store_true",
            default=False,
            help="Estimate area transition rate.")
    args = parser.parse_args()
    job_builder = validatelib.JobBuilder(
            nreps=args.nreps,
            random_seed=args.random_seed,
            job_queue=args.queue,
            virtual_env_path=args.venv,
            )

    ### PRODUCTION ###
    # num_focal_areas_regimes = [1, 8, ]
    # num_supplemental_areas_regimes = [0, 1, ]
    # birth_rate_regimes = [0.01, 0.05, 0.1,]
    # termination_condition_regime_types = ["target_focal_area_lineages"]
    # termination_condition_regime_values = [200,]

    # num_focal_areas_regimes = [1, ]
    # num_supplemental_areas_regimes = [0, ]
    # birth_rate_regimes = [0.1,]
    # termination_condition_regime_types = ["target_focal_area_lineages"]
    # termination_condition_regime_values = [50,]

    # ### DEVELOPMENT ###
    num_focal_areas_regimes = [2, 4,]
    num_supplemental_areas_regimes = [0, ]
    birth_rate_regimes = [0.0030, 0.0300]
    death_rate_regimes = [0.0,]
    # global_dispersal_rate_regimes = [0.00030, 0.0015, 0.0030, 0.006, 0.030]
    global_dispersal_rate_regimes = [
            0.0015, 0.0030, 0.0060,
            0.0150, 0.0300, 0.0600,
            0.1500, 0.3000,
            ]
    termination_condition_regime_types = ["target_focal_area_lineages"]
    termination_condition_regime_values = [200, ]
    regime_sources = [
            num_focal_areas_regimes,
            num_supplemental_areas_regimes,
            birth_rate_regimes,
            death_rate_regimes,
            global_dispersal_rate_regimes,
            termination_condition_regime_types,
            termination_condition_regime_values,
            ]
    regimes = itertools.product(*regime_sources)

    for regime_idx, regime in enumerate(regimes):
        output_prefix = "r{:05d}".format(regime_idx+1)
        num_focal_areas = regime[0]
        num_supplemental_areas = regime[1]
        birth_rate = regime[2]
        death_rate = regime[3]
        global_dispersal_rate = regime[4]
        termination_condition_type = regime[5]
        termination_condition_value = regime[6]

        simulator_model_kwargs = {
                "num_focal_areas": num_focal_areas,
                "num_supplemental_areas": num_supplemental_areas,
                "lineage_birth_rate": ("fixed_value", birth_rate, birth_rate),
                "lineage_death_rate": ("fixed_value", death_rate, death_rate),
                "global_dispersal_rate": global_dispersal_rate,
                "lineage_dispersal_weight": ("fixed_value", 1.0, 1.0),
                "termination_condition_type": termination_condition_type,
                "termination_condition_value": termination_condition_value,
        }
        profiler_options = [
                    "--no-estimate-trait-transition",
                    "-l", "birth.rate:{}".format(birth_rate),
                    "-l", "num.focal.areas:{}".format(num_focal_areas),
                    "-l", "num.supplemental.areas:{}".format(num_supplemental_areas),
                    "-l", "global.dispersal.rate:{}".format(global_dispersal_rate),
                    # "-l", "dispersal.rate:{}".format(global_dispersal_rate),
        ]
        if not args.estimate_dec_biogeobears and not args.estimate_dec_lagrange and not args.estimate_area_transition:
            raise ValueError("Neither dispersal rates nor area transition rates will be estimated")
        if not args.estimate_area_transition:
            profiler_options.append("--no-estimate-area-transition")
        if args.estimate_dec_biogeobears:
            profiler_options.append("--estimate-dec-biogeobears")
        if args.estimate_dec_lagrange:
            profiler_options.append("--estimate-dec-lagrange")
        job_builder.create_job(
                simulator_model_kwargs=simulator_model_kwargs,
                profiler_options=profiler_options,
                output_prefix=output_prefix,
                )

if __name__ == "__main__":
    main()

