#! /usr/bin/env python

import sys
import os
import json
import random
import argparse
import itertools
import collections

from archipelago import simulate

"""
## REFERENCE RANGES

-   areas:                        4, 8
-   birth rates:                  0.03,
-   extinction rates:             0.0, 0.03,
-   dispersal rates:              0.03, 0.15, 0.30
-   trait_transition rates:       0.03, 0.15, 0.30
-   taxa                          50

## WALLACEA

-   areas:                        4
-   birth rates:                  0.05482309568853121, 0.042834005131106676
-   extinction rates:             0.0, 0.00205547, 0.00257819,
-   dispersal rates:              0.0179814, 0.0143178
-   trait_transition rates:       0.062247, 0.047111, 43.815203, 0.002914
-   taxa                          198, 365

## WEST INDIES

-   areas:                        7
-   birth rates:                  0.05486829846551938, 0.034297575128763
-   extinction rates:             0.0, 8.82196e-05,
-   dispersal rates:              0.00616188, 0.00620009, 0.00424767
-   trait_transition rates:       0.011591, 0.047538, 0.015593, 0.231352, 0.018241, 0.001524, 0.003465,
-   taxa                          234, 76


"""

general_job_template = """\
#! /bin/bash
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -l h_vmem=12G
#$ -l virtual_free=12G
{queue}
{job_commands}
"""

def to_json(o):
    if isinstance(o, DynamicValue):
        return o.value
    raise TypeError(repr(o) + " is not JSON serializable")

class DynamicValue(object):

    def __init__(self, value=None):
        self.value = value

def generate_area_definitions(
        num_focal_areas,
        num_supplemental_areas=0):
    areas = []
    for i in range(num_supplemental_areas):
        areas.append({
            "label": "b{}".format(i+1),
            "is_supplemental": True,
            })
    for i in range(num_focal_areas):
        areas.append({
            "label": "a{}".format(i+1),
            "is_supplemental": False,
            })
    return areas

# def get_num_focal_areas_range():
#     return [4, 8]

# def get_num_supplemental_areas_range():
#     return [0, 1]

# def get_birth_rate_range():
#     reference = [0.03,]
#     reference.extend([0.05482309568853121, 0.042834005131106676]) # wallacean
#     reference.extend([0.05486829846551938, 0.034297575128763]) # west indian
#     return sorted(reference)

# def get_extinction_rate_range():
#     reference = [0,00, 0.03,]
#     reference.extend([0.0, 8.82196e-05,]) # wallacean
#     reference.extend([0.05486829846551938, 0.034297575128763]) # west indian
#     return sorted(reference)

# def get_dispersal_weight_range():
#     reference = [0.03, 0.15, 0.30]
#     reference.extend([0.0179814, 0.0143178,]) # wallacean
#     reference.extend([0.00616188, 0.00620009, 0.00424767]) # west indian
#     return sorted(reference)

# def get_trait_transition_rate_range():
#     reference = [0.03, 0.15, 0.30]
#     reference.extend([0.062247, 0.047111, 43.815203, 0.002914]) # wallacean
#     reference.extend([0.011591, 0.047538, 0.015593, 0.231352, 0.018241, 0.001524, 0.003465,]) # west indian
#     return sorted(reference)

# def get_target_focal_area_lineages_range():
#     reference = [50, ]
#     reference.extend([198, 365]) # wallacean
#     reference.extend([75, 234]) # west indian
#     return sorted(reference)

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-test-data-per-regime", "--ntest",
            type=int,
            default=100,
            help="Number of test data replicates per 'constrained' and 'unconstrained' regimes (default = %(default)s).")
    parser.add_argument("--num-training-data-per-regime", "--ntraining",
            type=int,
            default=100,
            help="Number of training data replicates per 'constrained' and 'unconstrained' regimes (default = %(default)s).")
    parser.add_argument(
            "--njob-replicates",
            "--num-replicates-per-configuration",
            type=int,
            default=1,
            help="Number of replicates per parameter combination (default = %(default)s).")
    parser.add_argument("--no-summary",
            action="store_true",
            default=False,
            help="Do not create summaries.")
    parser.add_argument("--profile", "--create-profile",
            dest="create_profile",
            action="store_true",
            default=False,
            help="Create profile (estimate birth, death, and dispersal rates on tree).")
    parser.add_argument("--no-profile-dispersal",
            action="store_true",
            default=False,
            help="Do not profile dispersal, even if profile is requested.")
    parser.add_argument("--venv",
            default=None,
            help="Path to Python virtual environment.")
    parser.add_argument("-z", "--random-seed",
            default=None,
            help="Seed for random number generator engine.")
    parser.add_argument("-q", "--queue",
            default=None,
            help="Name of queue to use (default: no queue specification")
    regime_options = parser.add_argument_group("Simulation Regime")
    regime_options.add_argument("--num-focal-areas",
            type=int,
            default=4,
            help="Number of focal areas (default: %(default)s).")
    regime_options.add_argument("--num-supplemental-areas",
            type=int,
            default=1,
            help="Number of supplemental areas (default: %(default)s).")
    regime_options.add_argument("--num-taxa",
            dest="target_focal_area_lineages",
            type=int,
            default=200,
            help="Number of taxa (default: %(default)s).")
    regime_options.add_argument("--pure-birth-only",
            action="store_const",
            const="pure-birth-only",
            dest="diversification_parameters",
            help="Suppress simulations where extinction rate is non-zero.")
    regime_options.add_argument("--no-pure-birth",
            action="store_const",
            const="no-pure-birth",
            dest="diversification_parameters",
            help="Suppress simulations where extinction rate is zero.")
    model_classification_options = parser.add_argument_group("Classification Options")
    model_classification_options.add_argument("--set-npca",
            default=None,
            metavar="n",
            help="Number of axes retained in the principal component analysis step.",
            )
    model_classification_options.add_argument("--optimize-npca",
            default=None,
            metavar="PENALTY-FACTOR",
            help=(
                "Use number of principal component axes that maximizes a "
                "score given by: (proportion of correct classifications  "
                "when reapplied to the training data) minus (the         "
                "proportion of available PC axes used multipled by the   "
                "given penalty factor). A penalty factor of 0.0 means to "
                "use the number of PC axes that maximizes the proportion "
                "of correct classifications. A penalty factor of 1.0     "
                "means that each additional axis retained will deduct    "
                "1.0/|PC| from the proportion of correct classification "
                "score."
                )
            )
    model_classification_options.add_argument("--maximize-npca",
            action="store_true",
            default=None,
            help="Retain all the principal component axes for the DAPC function.")

    args = parser.parse_args()

    if args.random_seed is None:
        args.random_seed = random.randint(0, sys.maxsize)
    rng = random.Random(args.random_seed)
    if args.venv is not None:
        venv_dir = os.path.expanduser(os.path.expandvars(args.venv))
        venv_activate = os.path.abspath(os.path.join(venv_dir, "bin", "activate"))
        if not os.path.exists(venv_activate):
            raise Exception("Virtual environment activation script not found: '{}'".format(venv_activate))
        source_venv = "source {}".format(venv_activate)
    else:
        source_venv = ""
    if args.queue:
        queue = "#$ -q {}".format(args.queue)
    else:
        queue = ""

    num_chosen = len(list(itertools.ifilter(lambda x: x is not None, [args.set_npca, args.optimize_npca, args.maximize_npca])))
    if num_chosen == 0:
        sys.exit("Must specify one of: '--set-npca', '--optimize-npca', or '--maximize-npca'")
    elif num_chosen > 1:
        sys.exit("Cannot specify more than one of: '--set-npca', '--optimize-npca', or '--maximize-npca'")
    elif args.set_npca is not None:
        npca = ["--set-npca", args.set_npca]
    elif args.optimize_npca is not None:
        npca = ["--optimize-npca", args.optimize_npca]
    elif args.maximize_npca is not None:
        npca = ["--maximize-npca"]

    lineage_birth_rate = DynamicValue()
    lineage_death_rate = DynamicValue()
    constrained_dispersal_weight = DynamicValue()
    unconstrained_dispersal_weight = DynamicValue()
    trait_transition_rate = DynamicValue()
    global_dispersal_rate = DynamicValue()
    target_focal_area_lineages = DynamicValue()

    base_model = {
        "diversification": {
            "lineage_birth_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_birth_rate,
            },
            "lineage_death_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_death_rate,
            },
        },
        "traits": [
            {"label": "habitat", "nstates": 2, "transition_rate": trait_transition_rate,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:habitat",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
        "termination_conditions": {"target_focal_area_lineages": target_focal_area_lineages},
    }


    # std_rng = [
    #         0.0015, 0.0030, 0.0060,
    #         0.0150, 0.0300, 0.0600,
    #         0.1500, 0.3000,
    #         ]
    std_rng = [
            0.0015, 0.0030, 0.0060,
            0.0150, 0.0300, 0.0600,
            0.1500, 0.3000,
            ]

    pure_birth_diversification_regimes = [
            (0.0015, 0.0),
            (0.0030, 0.0),
            (0.0060, 0.0),
            (0.0150, 0.0),
            (0.0300, 0.0),
            (0.0600, 0.0),
            ]

    non_pure_birth_diversification_regimes = [
            (0.0015, 0.0005),
            (0.0015, 0.0010),
            (0.0030, 0.0010),
            (0.0030, 0.0020),
            (0.0060, 0.0020),
            (0.0060, 0.0040),
            (0.0150, 0.0050),
            (0.0150, 0.0100),
            (0.0300, 0.0100),
            (0.0300, 0.0200),
            (0.0600, 0.0200),
            (0.0600, 0.0400),
            ]

    diversification_regimes = []
    if args.diversification_parameters != "no-pure-birth":
        diversification_regimes.extend(pure_birth_diversification_regimes)
    if args.diversification_parameters != "pure-birth-only":
        diversification_regimes.extend(non_pure_birth_diversification_regimes)
    if not args.no_summary:
        run_types = ("training", "target")
    else:
        run_types = ("run")
    global_dispersal_rate_regimes   = list(std_rng)
    trait_transition_rate_regimes   = list(std_rng)
    # target_focal_area_lineages_regimes = [50, 200,]
    regimes = (
            # num_focal_areas_regimes,
            # num_supplemental_areas_regimes,
            diversification_regimes,
            global_dispersal_rate_regimes,
            trait_transition_rate_regimes,
            # target_focal_area_lineages_regimes,
            )
    model_filepaths = []
    job_filepaths = []
    num_focal_areas_value = args.num_focal_areas
    num_supplemental_areas_value = args.num_supplemental_areas
    target_focal_area_lineages_value = args.target_focal_area_lineages
    param_config_index = 0
    for (
            # num_focal_areas_value,
            # num_supplemental_areas_value,
            birth_death_rate_value,
            global_dispersal_rate_value,
            trait_transition_rate_value,
            # target_focal_area_lineages_value,
            ) in itertools.product(*regimes):
        param_config_index += 1
        birth_rate_value = birth_death_rate_value[0]
        death_rate_value = birth_death_rate_value[1]
        lineage_birth_rate.value = birth_rate_value
        lineage_death_rate.value = death_rate_value
        trait_transition_rate.value = trait_transition_rate_value
        global_dispersal_rate.value = global_dispersal_rate_value
        target_focal_area_lineages.value = target_focal_area_lineages_value
        # subdir = "ntax{}_numareas{}+{}".format(target_focal_area_lineages_value, num_focal_areas_value, num_supplemental_areas_value)
        # if not os.path.exists(subdir):
        #     os.makedirs(subdir)
        for njob_idx, njob in enumerate(range(args.njob_replicates)):
            job_name = "a{:04d}.{:04d}".format(param_config_index, njob_idx+1)
            job_commands = []
            if source_venv:
                job_commands.append(source_venv)
            summary_paths = collections.defaultdict(dict)
            for is_constrained in (True, False):
                if is_constrained:
                    model_category = "constrained"
                    constrained_dispersal_weight.value = 0.0
                    unconstrained_dispersal_weight.value = 1.0 * 2
                else:
                    model_category = "unconstrained"
                    constrained_dispersal_weight.value = 1.0
                    unconstrained_dispersal_weight.value = 1.0
                configuration_name = "{}.{}".format(job_name, model_category)
                sim_model = {}
                sim_model["areas"] = generate_area_definitions(num_focal_areas=num_focal_areas_value, num_supplemental_areas=num_supplemental_areas_value)
                sim_model.update(base_model)
                # model_filepath = os.path.join(subdir, configuration_name + ".json")
                model_filepath = configuration_name + ".json"
                # sys.stderr.write("Model: {}\n".format(model_filepath))
                with open(model_filepath, "w") as mf:
                    json.dump(sim_model, mf, indent=4, default=to_json)
                model_filepaths.append(model_filepath)
                for run_type in run_types:
                    if run_type == "target":
                        nsim_reps = args.num_test_data_per_regime
                    else:
                        nsim_reps = args.num_training_data_per_regime
                    run_name = "{}.{}".format(configuration_name, run_type)
                    run_cmd = []
                    run_cmd.append("archipelago-simulate.py")
                    run_cmd.extend(["--nreps", str(nsim_reps)])
                    run_cmd.extend(["-z", str(rng.randint(0, sys.maxsize))])
                    run_cmd.extend(["--output-prefix", run_name])
                    run_cmd.extend(["--stderr-logging-level", "warning"])
                    run_cmd.append(os.path.basename(model_filepath))
                    run_cmd = " ".join(run_cmd)
                    job_commands.append(run_cmd)
                    result_trees_filepath = simulate.ArchipelagoSimulator.compose_focal_areas_trees_filepath(run_name)
                    if args.create_profile:
                        profile_cmd = []
                        profile_cmd.append("archipelago-profile-trees.py")
                        profile_cmd.extend(["--model-file", os.path.basename(model_filepath)])
                        profile_cmd.extend(["-l", "model.category:{}".format(model_category)])
                        profile_cmd.extend(["-l", "true.birth.rate:{}".format(birth_rate_value)])
                        profile_cmd.extend(["-l", "true.death.rate:{}".format(death_rate_value)])
                        profile_cmd.extend(["-l", "true.dispersal.rate:{}".format(global_dispersal_rate_value)])
                        profile_cmd.extend(["-l", "true.trait.transition.rate:{}".format(trait_transition_rate_value)])
                        if not args.no_profile_dispersal:
                            profile_cmd.append("--estimate-dec-biogeobears")
                        profile_cmd.append("--no-estimate-area")
                        profile_cmd.append("--ignore-estimation-errors")
                        profile_cmd.extend(["--verbosity", "0"])
                        profile_cmd.extend(["-o", "{}.profile.csv".format(run_name)])
                        profile_cmd.append(result_trees_filepath)
                        profile_cmd = " ".join(profile_cmd)
                        job_commands.append(profile_cmd)
                    if not args.no_summary:
                        summary_path = "{}.summary.csv".format(run_name)
                        summary_paths[run_type][model_category] = summary_path
                        summary_cmd = []
                        summary_cmd.append("archipelago-summarize.py")
                        # summary_cmd.extend(["-l", "dispersal.model:{}".format(model_category)])
                        summary_cmd.extend(["-l", "model.category:{}".format(model_category)])
                        summary_cmd.extend(["-l", "birth.rate:{}".format(birth_rate_value)])
                        summary_cmd.extend(["-l", "death.rate:{}".format(death_rate_value)])
                        summary_cmd.extend(["-l", "dispersal.rate:{}".format(global_dispersal_rate_value)])
                        summary_cmd.extend(["-l", "trait.transition.rate:{}".format(trait_transition_rate_value)])
                        summary_cmd.extend(["-l", "num.focal.areas:{}".format(num_focal_areas_value)])
                        summary_cmd.extend(["-l", "num.supplemental.areas:{}".format(num_supplemental_areas_value)])
                        summary_cmd.extend(["-l", "ntax:{}".format(target_focal_area_lineages_value)])
                        summary_cmd.extend(["-o", summary_path])
                        summary_cmd.append("-q")
                        summary_cmd.append(result_trees_filepath)
                        summary_cmd = " ".join(summary_cmd)
                        job_commands.append(summary_cmd)
            if not args.no_summary:
                model_classification_output_name = "{}.classification.csv".format(job_name)
                performance_output_name = "{}.performance.csv".format(job_name)
                for idx, true_model_category in enumerate(["constrained", "unconstrained"]):
                    performance_command = []
                    performance_command.append("archipelago-classify.py")
                    # performance_command.append("--optimize-npca")
                    # performance_command.extend(["--optimization-penalty-factor", "1.0"])
                    performance_command.extend(npca)
                    performance_command.extend(["-l", "true.model:{}".format(true_model_category)])
                    performance_command.extend(["-l", "birth.rate:{}".format(birth_rate_value)])
                    performance_command.extend(["-l", "death.rate:{}".format(death_rate_value)])
                    performance_command.extend(["-l", "dispersal.rate:{}".format(global_dispersal_rate_value)])
                    performance_command.extend(["-l", "trait.transition.rate:{}".format(trait_transition_rate_value)])
                    performance_command.extend(["-l", "num.focal.areas:{}".format(num_focal_areas_value)])
                    performance_command.extend(["-l", "num.supplemental.areas:{}".format(num_supplemental_areas_value)])
                    performance_command.extend(["-l", "ntax:{}".format(target_focal_area_lineages_value)])
                    performance_command.extend(["--true-model", true_model_category])
                    performance_command.extend(["--primary-output-filepath", model_classification_output_name])
                    performance_command.extend(["--performance-output-filepath", performance_output_name])
                    if idx > 0:
                        performance_command.append("--append")
                    performance_command.append(summary_paths["target"][true_model_category])
                    performance_command.append(summary_paths["training"]["constrained"])
                    performance_command.append(summary_paths["training"]["unconstrained"])
                    performance_command = " ".join(performance_command)
                    job_commands.append(performance_command)

            job_filename = "{}.job".format(job_name)
            job_filepath = job_filename
            sys.stderr.write("Job: {}\n".format(job_filepath))
            with open(job_filepath, "w") as jobf:
                template = general_job_template
                jobf.write(template.format(
                    job_commands="\n".join(job_commands),
                    queue=queue,
                    ))
            # job_filepaths.append((subdir, job_filename))
            job_filepaths.append(job_filename)
    assert len(model_filepaths) == len(set(model_filepaths))
    assert len(job_filepaths) == len(set(job_filepaths))
    with open("submit.sh", "w") as out:
        out.write("#! /bin/bash\n")
        for job_filepath in job_filepaths:
            out.write("qsub {}\n".format(job_filepath))
    sys.stderr.write("---\n")
    sys.stderr.write("{} models created\n".format(len(model_filepaths)))
    sys.stderr.write("{} jobs created\n".format(len(job_filepaths)))

if __name__ == "__main__":
    main()


