#! /usr/bin/env python

import os
import sys
import random
import argparse
import random
import glob
import collections
import re
from archipelago import simulate
from archipelago import profile

GENERAL_JOB_TEMPLATE = """\
#! /bin/bash
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -l h_vmem=12G
#$ -l virtual_free=12G
{queue}
{commands}
"""

MODEL_TEMPLATE = """\
{{
    "areas": {areas},
    "traits" : {traits},
    "diversification": {{
        "lineage_birth_rate": {lineage_birth_rate},
        "lineage_death_rate": {lineage_death_rate},
    }},
    "dispersal": {{
        "global_dispersal_rate" : {global_dispersal_rate},
        "lineage_dispersal_weight": {lineage_dispersal_weight},
    }},
    "termination_conditions": {{
        "{termination_condition_type}": {termination_condition_value},
    }}
}}

"""
def generate_areas_definition(
        num_focal_areas,
        num_supplemental_areas=0):
    areas = []
    for area_type_idx, area_count in enumerate((num_supplemental_areas, num_focal_areas,)):
        if area_type_idx == 0:
            is_supplemental = True
            prefix = "s"
        else:
            is_supplemental = False
            prefix = "a"
        for j in range(area_count):
            d = {}
            d["label"] = "{}{}".format(prefix, j+1)
            d["is_supplemental"] = is_supplemental
            areas.append(d)
    return areas

def generate_trait_definitions(label_nstates_rate_list):
    if not label_nstates_rate_list:
        return []
    s = []
    for label, nstates, rate in label_nstates_rate_list:
        s.append({
            "label": label,
            "nstates": nstates,
            "transition_rate": rate,
        })
    return s

def compose_rate_definition(definition_type, definition, description=None):
    d = {}
    d["definition_type"] = definition_type
    d["definition"] = definition
    d["description"] = description
    return d

def compose_model_definition(**kwargs):
    areas = generate_areas_definition(
        num_focal_areas=kwargs.pop("num_focal_areas"),
        num_supplemental_areas=kwargs.pop("num_supplemental_areas"))
    traits = generate_trait_definitions(kwargs.pop("traits", []))
    lineage_birth_rate = compose_rate_definition(
        *(kwargs.pop("lineage_birth_rate"))
    )
    lineage_death_rate = compose_rate_definition(
        *(kwargs.pop("lineage_death_rate"))
    )
    global_dispersal_rate = kwargs.pop("global_dispersal_rate")
    lineage_dispersal_weight = compose_rate_definition(
        *(kwargs.pop("lineage_dispersal_weight"))
    )
    termination_condition_type = kwargs.pop("termination_condition_type")
    termination_condition_value = kwargs.pop("termination_condition_value")
    if kwargs:
        raise ValueError("Unsupported keyword arguments: {}".format(kwargs))
    return MODEL_TEMPLATE.format(
            areas=areas,
            traits=traits,
            lineage_birth_rate=lineage_birth_rate,
            lineage_death_rate=lineage_death_rate,
            global_dispersal_rate = global_dispersal_rate,
            lineage_dispersal_weight=lineage_dispersal_weight,
            termination_condition_type=termination_condition_type,
            termination_condition_value=termination_condition_value,
    )

class JobBuilder(object):

    @staticmethod
    def get_general_validator_arg_parser():
        parser = argparse.ArgumentParser(
                add_help=False,
                )
        validator_options = parser.add_argument_group("Validator Options")
        validator_options.add_argument("--nreps",
                type=int,
                default=100,
                help="number of replicates (default = %(default)s).")
        validator_options.add_argument("-z", "--random-seed",
                default=None,
                help="Seed for random number generator engine.")
        validator_options.add_argument("--venv",
                default=None,
                help="Path to Python virtual environment.")
        validator_options.add_argument( "--queue",
                default=None,
                help="Name of queue to use (default: no queue)")
        return parser

    def __init__(self,
            nreps=None,
            random_seed=None,
            job_queue=None,
            virtual_env_path=None,
            ):
        if nreps is not None:
            self.nreps = nreps
        else:
            self.nreps = 100
        if random_seed is None:
            random_seed = random.randint(0, sys.maxsize)
        self.rng = random.Random(random_seed)
        if virtual_env_path is not None:
            self.venv_dir = os.path.expanduser(os.path.expandvars(args.venv))
            self.venv_activate = os.path.abspath(os.path.join(venv_dir, "bin", "activate"))
            if not os.path.exists(self.venv_activate):
                raise Exception("Virtual environment activation script not found: '{}'".format(self.venv_activate))
            self.source_venv_command = "source {}".format(venv_activate)
        else:
            self.source_venv_command = ""
        if job_queue:
            self.queue_command = "#$ -q {}".format(args.queue)
        else:
            self.queue_command = ""

    def create_job(self,
            simulator_model_kwargs,
            profiler_options,
            output_prefix,
            ):
        model_str = compose_model_definition(**simulator_model_kwargs)
        model_filepath = output_prefix + ".model.py"
        sys.stderr.write("Model: {}\n".format(model_filepath))
        with open(model_filepath, "w") as mf:
            mf.write(model_str)
        commands = []
        if self.source_venv_command:
            commands.append(self.source_venv_command)
        run_cmd = []
        run_cmd.append("archipelago-simulate.py")
        run_cmd.extend(["--nreps", str(self.nreps)])
        run_cmd.extend(["-z", str(self.rng.randint(0, sys.maxsize))])
        run_cmd.extend(["--output-prefix", output_prefix])
        run_cmd.extend(["--stderr-logging-level", "warning"])
        run_cmd.append(os.path.basename(model_filepath))
        run_cmd = " ".join(run_cmd)
        commands.append(run_cmd)
        analysis_cmd = []
        analysis_cmd.append("archipelago-profile-trees.py")
        analysis_cmd.extend(["--model-file", model_filepath])
        analysis_cmd.extend(["-v", "0"])
        analysis_cmd.extend(profiler_options)
        analysis_cmd.extend(["-o", "{}.profile.csv".format(output_prefix)])
        analysis_cmd.append(simulate.ArchipelagoSimulator.compose_focal_areas_trees_filepath(output_prefix))
        analysis_cmd = " ".join(analysis_cmd)
        commands.append(analysis_cmd)
        job_filepath = output_prefix + ".job"
        with open(job_filepath, "w") as jobf:
            jobf.write(GENERAL_JOB_TEMPLATE.format(
                commands="\n".join(commands),
                queue=self.queue_command,
                ))

def parse_fieldname_and_value(labels):
    if not labels:
        return collections.OrderedDict()
    fieldname_value_map = collections.OrderedDict()
    for label in labels:
        match = re.match(r"\s*(.*)\s*:\s*(.*)\s*", label)
        if not match:
            raise ValueError("Cannot parse fieldname and label (format required: fieldname:value): {}".format(label))
        fieldname, value = match.groups(0)
        fieldname_value_map[fieldname] = value
    return fieldname_value_map

def get_source_filepaths(dirs,
        model_file_schema="python",
        model_file_suffix=None,
        ignore_missing=False,
        ):
    assert model_file_schema in ("python", "json")
    model_profile_pairs = []
    if model_file_suffix is None:
        if model_file_schema == "python":
            model_file_suffix = ".model.py"
        elif model_file_schema == "json":
            model_file_suffix = ".model.json"
    for dirpath in dirs:
        for filepath in glob.glob(os.path.join(dirpath, "*" + model_file_suffix)):
            model_filepath = filepath
            stemname = model_filepath[:-len(model_file_suffix)]
            csv_filepath = stemname + ".profile.csv"
            if not os.path.exists(csv_filepath):
                if not ignore_missing:
                    raise Exception("Profile file for model '{}' not found: {}".format(model_filepath, csv_filepath))
            model_profile_pairs.append( (model_filepath, csv_filepath) )
    return model_profile_pairs

